from spheres import *
import uuid

########################################################################################

class jsCall:
    def __init__(self, view, name):
        self.view = view
        self.name = name
        
    def __call__(self, *args):
        data = []
        null = False
        again = False
        finished = False

        def __callback__(*returned):
            nonlocal data, null, again, finished
            returned = list(returned)
            if len(returned) == 1\
                    and type(returned[0]) == type({})\
                        and "error" in returned[0]:
                error = returned[0]["error"]
                if error.startswith("client attribute"):
                    null = True
                    finished = True
                    raise NameError("name '%s' of %s is not defined"\
                                     % (returned["attribute"],\
                                        type(self.view).__name__))
                elif error.startswith("client object"):
                    sockets.emit("create", {"class": self.view.__js_class__,\
                                            "uuid": self.view.uuid,\
                                            "args": {}})
                    again = True
                    finished = True
            else:
                data = returned
                finished = True
             
        sockets.emit("call", {"uuid": self.view.uuid,\
                              "func": self.name,\
                              "args": args},\
                              callback=__callback__)
        t = 0
        while not finished:
            t += 1
            if t > 200:
                finished = True
                raise Exception("Timed out!")
            sockets.sleep(0.01)
        if again:
            return self.__call__(*args)
        if not null:
            return data[0] if len(data) == 1 else data

########################################################################################

class View(object):
    views = {}
    expression_types = []
    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj, *args, **kwargs):
        object.__setattr__(self, "_obj", obj)
        self.uuid = str(uuid.uuid4())
        View.views[self.uuid] = self

        self.listeners = {}
        self.__outer_class__ = type(self).__name__[:type(self).__name__.index("(")]
        self.__inner_class__ = type(obj)
        self.__js_class__ = kwargs["js_class"]\
                                if "js_class" in kwargs\
                                    else self.__outer_class__
        self.__to_client__ = kwargs["to_client"]\
                                    if "to_client" in kwargs\
                                        else lambda view: str(view)
        self.__from_client__ = kwargs["from_client"]\
                                    if "from_client" in kwargs\
                                        else lambda data: object.__getattribute__(self, "_obj")
        self.requires_flush = ["__delitem__",  "__delslice__", "__setitem__", "__setslice__"]
        if "requires_flush" in kwargs:
            self.requires_flush.extend(kwargs["requires_flush"])

        self.local = kwargs["local"] if "local" in kwargs else False
        self.name = kwargs["name"] if "name" in kwargs else ""
   
    ########################################################################################

    @property
    def __class__(self):
        return self.__inner_class__

    def get(self):
        return object.__getattribute__(self, "_obj")

    def set(self, value, silent=False, local=False, exclude=[]):
        if value.__class__ in self.expression_types:
            value.__implement_set__(self, silent=silent)
            if not silent:
                jsCall(self, "refresh_from_server")(self.__to_client__(self))
        else:
            if type(value) == self.__inner_class__:
                object.__setattr__(self, "_obj", value)
                if not silent:
                    self.flush(local=local, exclude=exclude)
            else:
                raise Exception("can't set %s to %s!" % (self.__inner_class__.__name__,
                                                        type(value).__name__))

    def __lshift__(self, value):
        self.set(value)
        return self

    def set_from_view(self, view):
        self.set(view.get())

    def flush(self, local=False, exclude=[]):
        for uuid, listener_data in self.listeners.items():
            with_func = listener_data["with_func"]
            expression_type = listener_data["expression_type"]
            if uuid in View.views and uuid not in exclude and expression_type == None:
                View.views[uuid].set(with_func(self), exclude=[self.uuid])
        if not local and not self.local:
            jsCall(self, "refresh_from_server")(self.__to_client__(self))

    ########################################################################################

    def refresh_from_client(self, data):
        object.__setattr__(self, "_obj", self.__from_client__(data))
        self.flush(local=True)

    ########################################################################################

    def listen(self, to_whom, with_func, expression_type=None):
        to_whom.listeners[self.uuid] = {"with_func": with_func,\
                                        "expression_type": expression_type}

    def unlisten(self, to_whom):
        del to_whom.listeners[self.uuid]

    ########################################################################################

    def loop_for(self, n, func, rate=1/6, sleep=0.0001):
        if n != 0:
            for i in range(n):
                if i % int(1/rate) == 0:
                    self << func(self)
                else:
                    self.set(func(self), silent=True)
                sockets.sleep(sleep)

    ########################################################################################

    def js(self):
        print("workspace.views['%s']" % self.uuid)

    ########################################################################################
    
    def __getattr__(self, name):
        if hasattr(object.__getattribute__(self, "_obj"), name):
            attribute = getattr(object.__getattribute__(self, "_obj"), name)
            if callable(attribute):
                def __wrapper__(*args, **kwargs):
                    args = [object.__getattribute__(arg, "_obj")\
                            if "View" in type(arg).__name__\
                            else arg\
                                for arg in args]
                    kwargs = dict([(key, object.__getattribute__(val, "_obj")\
                            if "View" in type(val).__name__\
                            else val ) for key, val in kwargs.items()])
                    value = attribute(*args)
                    self.flush() if name in self.requires_flush else None
                    return value
                return __wrapper__
            else:
                self.flush() if name in self.requires_flush else None
                return attribute
        else:
            return jsCall(self, name)

    def __delattr__(self, name):
        if hasattr(object.__getattribute__(self, "_obj"), name):
            delattr(object.__getattribute__(self, "_obj"), name)
            self.flush()
        else:
            super().__delattr__(name)

    def __setattr__(self, name, value):
        if hasattr(object.__getattribute__(self, "_obj"), name):
            setattr(object.__getattribute__(self, "_obj"), name, value)
            self.flush()
        else:
            super().__setattr__(name, value)

    ########################################################################################

    def __del__(self):
        sockets.emit("destroy", {"uuid": self.uuid})
    
    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        self.flush()
        return repr(object.__getattribute__(self, "_obj"))

    ########################################################################################

    _special_names = [
        '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', 
        '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__', 
        '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__', 
        '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
        '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__', 
        '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__', 
        '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', 
        '__long__', '__lt__', '__mod__', '__mul__', '__ne__', #'__lshift__',
        '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__', 
        '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__', 
        '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__', #'__repr__',
        '__rmul__', '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', 
        '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__', 
        '__truediv__', '__xor__', 'next',
    ]

    @classmethod
    def _create_class_proxy(cls, inner_class):        
        def make_method(name):
            def method(self, *args, **kwargs):
                args = [object.__getattribute__(arg, "_obj")\
                            if "View" in type(arg).__name__\
                            else arg\
                                for arg in args]
                kwargs = dict([(key, object.__getattribute__(val, "_obj")\
                            if "View" in type(val).__name__\
                            else val ) for key, val in kwargs.items()])
                value = getattr(object.__getattribute__(self, "_obj"), name)(*args, **kwargs)
                self.flush() if name in self.requires_flush else None
                return value
            return method
        namespace = {}
        for name in cls._special_names:
            if hasattr(inner_class, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, inner_class.__name__), (cls,), namespace)
    
    def __new__(cls, obj, *args, **kwargs):
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            inner_class = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = inner_class = cls._create_class_proxy(obj.__class__)
        inself = object.__new__(inner_class)
        inner_class.__init__(inself, obj, *args, **kwargs)
        return inself

    @classmethod
    def register_expression_type(cls, expression_type):
        View.expression_types.append(expression_type)

    @classmethod
    def unregister_expression_type(cls, expression_type):
        del View.expression_types[expression_type]

########################################################################################



import math
import sympy
import numpy as np
import functools
import qutip as qt
import itertools
import scipy

####################################################################

def d_j(d):
    return (d-1)/2

def j_d(j):
    return 2*j + 1

####################################################################

def normalize(v):
    n = np.linalg.norm(v)
    if n != 0:
        return v/n
    else:
        return v

def get_phase(v):
    c = None
    if isinstance(v, qt.Qobj):
        v = v.full().T[0]
    i = (v!=0).argmax(axis=0)
    c = v[i]
    return np.exp(1j*np.angle(c))

def normalize_phase(v):
    return v/get_phase(v)

####################################################################

def random_complex_unit_vector(d):
	v = np.random.randn(d) + 1j*np.random.randn(d)
	return v/np.linalg.norm(v)

def random_hermitian_matrix(d):
	H = np.random.randn(d,d) + 1j*np.random.randn(d, d)
	return H + H.T.conj()

####################################################################

def c_xyz(c, pole="south"):
    if(pole == "south"):
        if c == float("Inf"):
            return np.array([0,0,-1])
        else:
            x, y = c.real, c.imag
            return np.array([2*x/(1 + x**2 + y**2),\
                             2*y/(1 + x**2 + y**2),\
                   (1-x**2-y**2)/(1 + x**2 + y**2)])
    elif (pole == "north"):
        if c == float("Inf"):
            return np.array([0,0,1])
        else:
            x, y = c.real, c.imag
            return np.array([2*x/(1 + x**2 + y**2),\
                             2*y/(1 + x**2 + y**2),\
                   (-1+x**2+y**2)/(1 + x**2 + y**2)])

def xyz_c(xyz, pole="south"):
    x, y, z = xyz
    if (pole=="south"):
        if np.isclose(z,-1):
            return float("Inf")
        else:
            return x/(1+z) + 1j*y/(1+z)
    elif (pole=="north"):
        if np.isclose(z,1):
            return float("Inf")
        else:
            return x/(1-z) + 1j*y/(1-z)

####################################################################

def spinor_xyz(spinor):
    if isinstance(spinor, np.ndarray):
        spinor = qt.Qobj(spinor)
    return np.array([qt.expect(qt.sigmax(), spinor),\
                     qt.expect(qt.sigmay(), spinor),\
                     qt.expect(qt.sigmaz(), spinor)])

def xyz_spinor(xyz):
    return c_spinor(xyz_c(xyz))

def spin_xyz(spin):
    if isinstance(spinor, np.ndarray):
        spinor = qt.Qobj(spinor)
    j = d_j(spin.shape[0])
    return np.array([qt.expect(qt.jmat(j, 'x'), spinor),\
                     qt.expect(qt.jmat(j, 'y'), spinor),\
                     qt.expect(qt.jmat(j, 'z'), spinor)])

####################################################################

def spinor_c(spinor):
    a, b = None, None
    if isinstance(spinor, qt.Qobj):
        a, b = spinor.full().T[0]
    else:
        a, b = spinor
    if np.isclose(a,0):
        return float('Inf')
    else:
        return b/a

def c_spinor(c):
    if c == float('Inf'):
        return np.array([0,1])
    else:
        return normalize(np.array([1, c]))

####################################################################

def spin_poly(spin):
    if isinstance(spin, qt.Qobj):
        spin = spin.full().T[0]
    j = (spin.shape[0]-1)/2.
    v = spin
    poly = []
    for m in np.arange(-j, j+1, 1):
        i = int(m+j)
        poly.append(v[i]*\
            (((-1)**(i))*math.sqrt(math.factorial(2*j)/\
                        (math.factorial(j-m)*math.factorial(j+m)))))
    return poly

def poly_spin(poly):
    j = (len(poly)-1)/2.
    spin = []
    for m in np.arange(-j, j+1):
        i = int(m+j)
        spin.append(poly[i]/\
            (((-1)**(i))*math.sqrt(math.factorial(2*j)/\
                        (math.factorial(j-m)*math.factorial(j+m)))))
    aspin = np.array(spin)
    return aspin/np.linalg.norm(aspin)

####################################################################

def poly_roots(poly):
    head_zeros = 0
    for c in poly:
        if c == 0:
            head_zeros += 1 
        else:
            break
    return [float("Inf")]*head_zeros + [complex(root) for root in np.roots(poly)]

def roots_coeffs(roots):
    n = len(roots)
    coeffs = np.array([((-1)**(-i))*sum([np.prod(term) for term in itertools.combinations(roots, i)]) for i in range(0, len(roots)+1)])
    return coeffs/coeffs[0]
        
def roots_poly(roots):
    zeros = roots.count(0j)
    if zeros == len(roots):
        return [1j] + [0j]*len(roots)
    poles = roots.count(float("Inf"))
    roots = [root for root in roots if root != float('Inf')]
    if len(roots) == 0:
        return [0j]*poles + [1j]
    #Z = sympy.symbols("Z")
    #Poly = sympy.Poly(functools.reduce(lambda a, b: a*b, [Z-root for root in roots]), domain="CC")
    #return [0j]*poles + [complex(c) for c in Poly.all_coeffs()] + [0j]*zeros
    return [0j]*poles + roots_coeffs(roots).tolist() #+ [0j]*(zeros-1)

####################################################################

def spin_roots(spin):
    if isinstance(spin, qt.Qobj):
        spin = spin.full().T[0]
    return poly_roots(spin_poly(spin))

def roots_spin(roots):
    return poly_spin(roots_poly(roots))

####################################################################

def spin_XYZ(spin):
    if isinstance(spin, qt.Qobj):
        spin = spin.full().T[0]
    return [c_xyz(root).tolist() for root in poly_roots(spin_poly(spin))]

def XYZ_spin(XYZ):
    #XYZ = [np.array(xyz) for xyz in XYZ]
    #XYZ = [xyz/np.linalg.norm(xyz) for xyz in XYZ]
    return qt.Qobj(poly_spin(roots_poly([xyz_c(xyz) for xyz in XYZ])))

####################################################################

def differentiate_spin(spin, times=1):
    def _differentiate_spin_(spin):
        n = len(spin)-1
        for i in range(len(spin)):
            spin[i] *= n
            n -= 1;
        return normalize(spin[:-1])
    spin = spin.copy()
    for t in range(times):
        spin = _differentiate_spin_(spin)
    return spin

def integrate_spin(spin, times=1):
    def _integrate_spin_(spin):
        return np.array(spin.tolist()+[0j])
    spin = spin.copy()
    for t in range(times):
        spin = _integrate_spin_(spin)
    return spin    

####################################################################

def sph_xyz(sph):
    r, theta, phi = sph;
    x = r*np.sin(theta)*np.cos(phi)
    y = r*np.sin(theta)*np.sin(phi)
    z = r*np.cos(theta)
    return np.array([x,y,z])

def xyz_sph(xyz):
    x, y, z = xyz
    r =  np.sqrt(x*x + y*y + z*z)
    theta =  np.arccos(z/r)
    phi =  np.arctan2(y,x)
    return np.array([r,theta,phi])

####################################################################

def eval_spin_at_c(spin, at):
    if at == float("Inf"):
        return eval_spin_at_c(XYZ_spin([c_xyz(xyz_c(xyz, pole="north")) for xyz in spin_XYZ(spin)]), 0)
    if isinstance(spin, qt.Qobj):
        spin = spin.full().T[0]
    normalized = spin/spin[0]
    poly = spin_poly(normalized)
    n = len(poly)
    return sum([poly[i]*(np.power(at,(n-i-1))) for i in range(n)])

def eval_spin_at(spin, at):
    return eval_spin_at_c(spin, xyz_c(at))

def eval_spin(spin, n=50, a=0.6):
    n=20
    theta, phi = np.linspace(0, 2*np.pi, 2*n), np.linspace(0, np.pi, n/2)
    points = []
    for t in theta:
        for p in phi:
            xyz = sph_xyz(np.array([1,t,p]))
            c = eval_spin_at(spin, xyz)
            points.append([xyz.tolist(), [c.real, c.imag], np.abs(c), np.angle(c)/(2*np.pi), c_xyz(c).tolist(), [((np.angle(c)/(np.pi))+1)/2, (1-np.power(a, np.abs(c))), 0.5]])
    return points

####################################################################

def husimi(spin, at):
    if isinstance(spin, qt.Qobj):
        spin = spin.full().T[0]
    n_stars = int(2*d_j(len(spin)))
    coherent = XYZ_spin([at]*n_stars)
    return np.vdot(spin, coherent)

def coherent(at, n):
    return XYZ_spin([at]*n)

def husimi_snapshot(spin, n=30):
    theta, phi = np.linspace(0, 2*np.pi, 16), np.linspace(0, np.pi, 16)
    points = []
    for t in theta:
        points_row = []
        for p in phi:
            #points.append([xyz.tolist(), [c.real, c.imag], np.abs(c), np.angle(c)/(2*np.pi), c_xyz(c).tolist(), [((np.angle(c)/(np.pi))+1)/2, (1-np.power(a, np.abs(c))), 0.5]])
            xyz = sph_xyz(np.array([1,t,p]))           
            h = husimi(spin, xyz)
            a = np.angle(h)
            points_row.append([xyz.tolist(), [np.abs(h), [np.cos(a), np.sin(a)]]])
        points.append(points_row)
    return points




import os
import sys
import json
import flask
import getopt
import logging
import socketio
import eventlet
import threading
import webbrowser
import eventlet.wsgi
import forbiddenfruit
from termcolor import colored

########################################################################################

sockets = socketio.Server(async_mode='threading')
app = flask.Flask(__name__)
app.wsgi_app = socketio.Middleware(sockets, app.wsgi_app)
log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)

########################################################################################

def show(s):
    print(colored("\n> ", "yellow")+colored(s, "magenta"))

def show_prelude():
    os.system("clear")
    print(colored("WELCOME", "green", "on_blue", attrs=['blink', 'bold'])+\
          colored(" TO ", "green", "on_blue", attrs=['blink', 'bold'])+\
          colored("SPHERES!", "green", "on_blue", attrs=['blink', 'bold']))

########################################################################################

from spheres.magic import *
from spheres.view import *
from spheres.expressions import *
from spheres.sphere import *

########################################################################################

@app.route("/")
def root():
    return flask.render_template("index.html")

########################################################################################

connected = False
@sockets.on("connect")
def connect(sid, data):
    global connected
    connected = True
    show("%s connected" % sid)

@sockets.on("disconnect")
def disconnect(sid):
    show("%s disconnected" % sid)

@sockets.on("call")
def call(sid, data):
    if data["uuid"] in View.views:
        obj = View.views[data["uuid"]]
        if hasattr(obj, data["func"]):
            return getattr(obj, data["func"])(*data["args"])
        else:
            return {"error": "server attribute %s not found!" % data["func"]}
    else:
        return {"error": "server object %s not found!" % data["uuid"]}

########################################################################################

def __init__(app, sockets, port=8080):
    show_prelude()
    app.run(threaded=True, port=port)

PORT = 8080
try:
    arguments, values = getopt.getopt(sys.argv[1:], "p:", ["port="])
    for arg, val in arguments:
        if arg in ("-p", "--port"):
            PORT = val
except getopt.error as err:
    print(str(err))
    sys.exit(2)

thread = threading.Thread(target=__init__, args=(app, sockets, PORT))
thread.start()

webbrowser.open_new_tab("http://localhost:8080")

########################################################################################

while not connected:
    pass


from spheres import *

def Sphere(obj, *args, **kwargs):
    kwargs = {**kwargs, \
              **{"to_client": lambda view:\
                    {"stars": spin_XYZ(view),\
                     "phase": [1,0]},\
                "from_client": lambda data: XYZ_spin(data["stars"]),\
                "js_class": "Sphere"}}
    return View(obj, *args, **kwargs)

def partials(pure):
    _partials = []
    n = len(pure.dims[0])
    for i in range(n):
        partial = View(pure.ptrace(i))
        def __pure_to_partial__(index, dims):
            def __pure_to_partial__inner__(_pure):
                val = _pure.get()
                val.dims = dims
                return val.ptrace(index)
            return __pure_to_partial__inner__
        partial.listen(pure, __pure_to_partial__(i, pure.dims))
        def __partial_to_pure__(index, dims):
            def __partial_to_pure__iner__(_partial, _pure, data):
                return qt.tensor(*[qt.identity(d)\
                    if j!=index else data["after"]\
                        for j, d in enumerate(dims[0])])*_pure.get()
            return __partial_to_pure__iner__
        pure.listen(partial, __partial_to_pure__(i, pure.dims),\
            expression_type=OperatorExpression)
        _partials.append(partial)
    return _partials


from spheres import *
import copy
import qutip as qt

########################################################################################

class OperatorExpression:
    def __init__(self, ops):
        self.ops = ops        

    def __mul__(self, other):
        if type(other) == OperatorExpression:
            return OperatorExpression(self.ops+other.ops)
        elif other.__class__ == qt.Qobj and other.type == "oper":
            return OperatorExpression(self.ops+[other])
        else:
            return self.reduced()*other

    def __rmul__(self, other):
        if type(other) == OperatorExpression:
            return OperatorExpression(other.ops+self.ops)
        elif other.__class__ == qt.Qobj and other.type == "oper":
            return OperatorExpression([other]+self.ops)
        else:
            return other*self.reduced()

    def __getattr__(self, name):
        return getattr(self.reduced(), name)

    def reduced(self, i=None):
        if i == None:
            return self.prod(self.ops)
        else:
            return self.prod(self.ops[:i]),\
                   self.prod(self.ops[i+1:]),\

    def prod(self, l):
        p = l[0]
        for el in l[1:]:
            p = View(p.__old_mul__(el), local=True)
        return p

    def __implement_set__(self, view, silent=False):
        object.__setattr__(view, "_obj", self.reduced())
        if not silent:
            for i, op in enumerate(self.ops):
                if "View" in type(op).__name__ and op.uuid == view.uuid:
                    before, after = self.reduced(i=i)
                    for uuid, listener_data in view.listeners.items():
                        with_func = listener_data["with_func"]
                        if uuid in View.views:
                            View.views[uuid].set(\
                                    with_func(view,\
                                    View.views[uuid],\
                                    {"before": before, "after": after}),\
                                    exclude=[view.uuid])

    def __repr__(self):
        return str(self.ops)

########################################################################################

_qview_type = type(View(qt.basis(2,0)))
setattr(_qview_type, "__old_mul__", copy.copy(_qview_type.__mul__))
setattr(_qview_type, "__old_rmul__", copy.copy(_qview_type.__rmul__))
def __new_mul__(self, other):
    if self.type == "bra":
        if type(other) == OperatorExpression:
            return self.__old_mul__(other.reduced())
    elif self.type == "oper":
        if type(other) == OperatorExpression:
            return OperatorExpression([self]+other.ops)
        elif other.__class__ == qt.Qobj and other.type == "oper":
            return OperatorExpression([self, other])
    return self.__old_mul__(other)
setattr(_qview_type, "__mul__", __new_mul__)
def __new_rmul__(self, other):
    if self.type == "ket":
        if type(other) == OperatorExpression:
            return other.reduced().__old_mul__(self)
    elif self.type == "oper":
        if type(other) == OperatorExpression:
            return OperatorExpression(other.ops+[self])
        elif other.__class__ == qt.Qobj and other.type == "oper":
            return OperatorExpression([other, self])
    return other.__old_mul__(self)
setattr(_qview_type, "__rmul__", __new_rmul__)
setattr(_qview_type, "__old_repr__", copy.copy(_qview_type.__repr__))
def __new_repr__(self):
    if self.name == "":
        return self.__old_repr__()
    self.__old_repr__()
    return self.name
setattr(_qview_type, "__repr__", __new_repr__)
setattr(_qview_type, "name", "")

_special_names = [
    '__abs__', '__add__', '__and__', '__call__', '__cmp__', '__coerce__', 
    '__contains__', '__delitem__', '__delslice__', '__div__', '__divmod__', 
    '__eq__', '__float__', '__floordiv__', '__ge__', '__getitem__', 
    '__getslice__', '__gt__', '__hash__', '__hex__', '__iadd__', '__iand__',
    '__idiv__', '__idivmod__', '__ifloordiv__', '__ilshift__', '__imod__', 
    '__imul__', '__int__', '__invert__', '__ior__', '__ipow__', '__irshift__', 
    '__isub__', '__iter__', '__itruediv__', '__ixor__', '__le__', '__len__', 
    '__long__', '__lshift__', '__lt__', '__mod__',  '__ne__', #'__mul__',
    '__neg__', '__oct__', '__or__', '__pos__', '__pow__', '__radd__', 
    '__rand__', '__rdiv__', '__rdivmod__', '__reduce__', '__reduce_ex__', #'__repr__', 
    '__reversed__', '__rfloorfiv__', '__rlshift__', '__rmod__',
    '__ror__', '__rpow__', '__rrshift__', '__rshift__', '__rsub__', #'__rmul__',
    '__rtruediv__', '__rxor__', '__setitem__', '__setslice__', '__sub__', 
    '__truediv__', '__xor__', 'next',
]
setattr(_qview_type, "__old_dag__", copy.copy(qt.Qobj.dag))
def __new_dag__(self):
    return View(self.__old_dag__(), local=True)
setattr(_qview_type, "dag", __new_dag__)

for name in _special_names:
    if name in dir(qt.Qobj):
        def wrap(name):
            def __wrapper__(self, *args, **kwargs):
                args = [arg.reduced() if type(arg) == OperatorExpression else arg for arg in args]
                kwargs = dict([(k, v.reduced() if type(v) == OperatorExpression else v) for k, v in kwargs.items()])
                return getattr(self.reduced(), name)(*args, **kwargs)
            return __wrapper__
        setattr(OperatorExpression, name, wrap(name))

########################################################################################

View.register_expression_type(OperatorExpression)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="spheres", 
    version="0.2.0.2",
    author="Matthew Weiss",
    author_email="heyredhat@gmail.com",
    description="functorial visualization library with support for quantum mechanics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/heyredhat/spheres",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    setup_requires=["numpy", "scipy", "cython"],\
    install_requires=["flask", "python-socketio", "eventlet",\
        "termcolor", "sympy", "numpy", "scipy", "cython", "qutip", "forbiddenfruit"],\
    python_requires=">=3",\
    package_data={
        "": ["templates/*",\
             "static/css/*",\
             "static/fonts/*",\
             "static/img/*",\
             "static/models/*",\
             "static/js/spheres/*",\
             "static/js/*.js",\
             "static/js/jquery-ui/*",\
             "static/js/three.js/build/three.min.js",\
             "static/js/three.js/examples/js/controls/DragControls.js",\
             "static/js/three.js/examples/js/controls/TransformControls.js",\
             "static/js/three.js/examples/js/utils/SceneUtils.js",\
             "static/js/three.js/examples/js/controls/OrbitControls.js",\
             "static/js/three.js/examples/js/controls/TrackballControls.js",\
             "static/js/three.js/examples/js/effects/OutlineEffect.js",\
             "static/js/three.js/examples/js/shaders/CopyShader.js",\
             "static/js/three.js/examples/js/shaders/AfterimageShader.js",\
             "static/js/three.js/examples/js/postprocessing/EffectComposer.js",\
             "static/js/three.js/examples/js/postprocessing/RenderPass.js",\
             "static/js/three.js/examples/js/shaders/HalftoneShader.js",\
             "static/js/three.js/examples/js/postprocessing/ShaderPass.js",\
             "static/js/three.js/examples/js/shaders/PixelShader.js",\
             "static/js/three.js/examples/js/postprocessing/AfterImagePass.js",\
             "static/js/three.js/examples/js/postprocessing/OutlinePass.js",\
             "static/js/three.js/examples/js/postprocessing/HalftonePass.js",\
             "static/js/three.js/examples/js/loaders/GLTFLoader.js"]\
    }
)